﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ACO
{
    public class AlgTypeACO : AlgType
    {
        #region fields and attributes

        public Boolean Elitism;
        public Boolean FootStepping;
        public Boolean MinMax;
        public Boolean System;

        #endregion

        #region constructors

        /// <summary>
        /// default constructor, sets all algorithm aditions to false
        /// </summary>
        public AlgTypeACO()
        {
            this.Elitism = false;
            this.FootStepping = false;
            this.MinMax = false;
            this.System = false;
        }

        /// <summary>
        /// full constructor version for one line creation and use
        /// </summary>
        /// <param name="elitism">defines to use elitism or not</param>
        /// <param name="footStepping">defines to use foot stepping or not</param>
        /// <param name="greedy">defines a greedy algorithm (overwrites all other additions)</param>
        /// <param name="minMax">defines to use min-max or not</param>
        /// <param name="system">defines to use Ant System pheromone adjustment or not</param>
        public AlgTypeACO(Boolean elitism, Boolean footStepping, Boolean minMax, Boolean system)
        {
            this.Elitism = elitism;
            this.FootStepping = footStepping;
            this.MinMax = minMax;
            this.System = system;
        }

        #endregion

        #region methods

        /// <summary>
        /// gets the algorithm type as a string for saving optimization data in files
        /// </summary>
        /// <returns>algorithm type as string for saving</returns>
        public override string AsString()
        {
            string s = "";
            if (Elitism)
            {
                s += "Elitism";
            }
            if (FootStepping)
            {
                if (s != "")
                {
                    s += ", ";
                }
                s += "FootStepping";
            }
            if (MinMax)
            {
                if (s != "")
                {
                    s += ", ";
                }
                s += "MinMax";
            }
            if (System)
            {
                if (s != "")
                {
                    s += ", ";
                }
                s += "System";
            }
            return s;
        }

        /// <summary>
        /// compares two ants looking only their route length, used to sort the ant list for elitism
        /// </summary>
        /// <param name="x">first ant</param>
        /// <param name="y">second ant</param>
        /// <returns>comparison result between the ants' route lengths</returns>
        private static int CompareAnts(Ant x, Ant y)
        {
            return x.foundRouteLength.CompareTo(y.foundRouteLength);
        }

        /// <summary>
        /// does algorithm extra-s depending on the chosen algorithm type,
        /// is called from ant colony on the end of every cycle
        /// </summary>
        internal override void DoVersionExtras()
        {
            if (this.System)
            {
                if (AntColony.Instance.cyclusCounter==1)
                {
                    AntColony.Instance.SetTrail(AntColony.Instance.ants[0].foundRouteLength);
                }
            }
            if (this.Elitism)
            {
                AntColony.Instance.ants.Sort(CompareAnts);
                int i = 1;
                foreach (Ant a in AntColony.Instance.ants)
                {
                    a.AddTrail(1 / i);
                    i++;
                    if (i > 10)
                    {
                        break;
                    }
                }
            }
            else
            {
                AntColony.Instance.AddTrail();
            }
            if (this.FootStepping)
            {
                for (int i = 0; i < AntColony.Instance.cityList.Count / 4.0; i++)
                {
                    int city = AntColony.Instance.randomizer.Next(AntColony.Instance.cityList.Count);
                    int maxTrailIndex = -1;
                    double maxTrail=0;
                    for (int j = 0; j < AntColony.Instance.cityList.Count; j++)
                    {
                        if (AntColony.Instance.trail[city, j] > maxTrail)
                        {
                            maxTrail = AntColony.Instance.trail[city, j];
                            maxTrailIndex = j;
                        }
                    }
                    AntColony.Instance.trail[city, maxTrailIndex] *= 1.05;
                    AntColony.Instance.trail[maxTrailIndex, city] *= 1.05;
                }
            }
            if (this.MinMax)
            {
                double sum=0;
                for (int i = 0; i<AntColony.Instance.cityList.Count;i++)
                {
                    for (int j = 0; j<AntColony.Instance.cityList.Count;j++)
                    {
                        sum+= AntColony.Instance.trail[i,j];
                    }
                }
                double min =(1.0/5.0) * (sum / (AntColony.Instance.cityList.Count*AntColony.Instance.cityList.Count));
                double max = 15*min;
                for (int i = 0; i<AntColony.Instance.cityList.Count;i++)
                {
                    for (int j = 0; j<AntColony.Instance.cityList.Count;j++)
                    {
                        if (AntColony.Instance.trail[i,j]<min)
                        {
                            AntColony.Instance.trail[i,j]=min;
                        }
                        else
                        {
                            if (AntColony.Instance.trail[i,j]>max)
                            {
                                AntColony.Instance.trail[i,j]=max;
                            }
                        }
                    }
                }
            }
            return;
        }

        #endregion
    }
}
