﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ACO
{
    internal class Ant
    {
        #region Fields

        internal List<int> tabuList;
        private int citysPerAgent;
        private int cityOverflowVRP;
        internal double foundRouteLength;

        #endregion

        #region Methods

        #region sets and resets

        /// <summary>
        /// constructor, initialises the tabu list
        /// </summary>
        internal Ant()
        {
            this.tabuList = new List<int>();
        }

        /// <summary>
        /// resets the ant (tabu list and route length)
        /// </summary>
        internal void ResetAnt()
        {
            this.tabuList.Clear();
            this.foundRouteLength = 0;
            return;
        }

        #endregion

        #region movement

        /// <summary>
        /// is called from ant colony once every cycle,
        /// calls "makeStep" as many times as there are citys in the loaded problem
        /// </summary>
        internal void makeCycle()
        {
            citysPerAgent = (int)Math.Floor((double)AntColony.Instance.cityList.Count / (double)AntColony.Instance.AgentNumber);
            cityOverflowVRP = AntColony.Instance.cityList.Count - citysPerAgent * AntColony.Instance.AgentNumber;
            for (int i = 0; i < AntColony.Instance.cityList.Count+1; i++)
            {
                this.makeStep();
            }
        }

        /// <summary>
        /// makes one step (moves to the next city)
        /// </summary>
        private void makeStep()
        {
            if (this.tabuList.Count == 0)
            {
                if (AntColony.Instance.ProblemTyp == ProblemType.VRP)
                {
                    this.tabuList.Add(0);
                }
                else
                {
                    this.tabuList.Add(AntColony.Instance.randomizer.Next(AntColony.Instance.cityList.Count));
                }
            }
            else
            {
                if (AntColony.Instance.ProblemTyp == ProblemType.VRP)
                {
                    this.ReturnHomeForNextAgent();
                }
                if (AntColony.Instance.AlgType.AsString() == "Greedy")
                {
                    #region greedy
                    double shortest = double.MaxValue;
                    int shortestLocation = -1;
                    for (int i = 0; i < AntColony.Instance.cityList.Count; i++)
                    {
                        if (!this.tabuList.Contains(i))
                        {
                            if (AntColony.Instance.distances[i, this.tabuList[tabuList.Count - 1]] < shortest)
                            {
                                shortest = AntColony.Instance.distances[i, this.tabuList[tabuList.Count - 1]];
                                shortestLocation = i;
                            }
                        }
                    }
                    if (shortestLocation != -1)
                    {
                        this.tabuList.Add(shortestLocation);
                    }
                    else
                    {
                        this.tabuList.Add(tabuList[0]);
                    }
                    #endregion
                }
                else
                {
                    #region probability sumation
                    List<double> probability = new List<double>();
                    double sum = 0;
                    for (int i = 0; i < AntColony.Instance.cityList.Count; i++)
                    {
                        if (this.tabuList.Contains(i))
                        {
                            probability.Add(0);
                        }
                        else
                        {
                            probability.Add(Math.Pow(AntColony.Instance.trail[this.tabuList[tabuList.Count - 1], i], AntColony.Instance.Alpha) *
                                            Math.Pow(AntColony.Instance.visibility[this.tabuList[tabuList.Count - 1], i], AntColony.Instance.Beta));
                            sum += probability[probability.Count - 1];
                        }
                    }
                    for (int i = 0; i < AntColony.Instance.cityList.Count; i++)
                    {
                        probability[i] = probability[i] / sum;
                    }
                    #endregion

                    #region Odi U Grad
                    if (sum > 0)
                    {
                        double random = AntColony.Instance.randomizer.NextDouble();
                        if (random == 0)
                        {
                            random = 1;
                        }
                        int placeToGo = -1;
                        while (random > 0)
                        {
                            placeToGo++;
                            random -= probability[placeToGo];
                        }
                        this.tabuList.Add(placeToGo);
                        return;
                    }
                    else
                    {
                        if (this.tabuList[this.tabuList.Count - 1] != this.tabuList[0])
                        {
                            this.tabuList.Add(this.tabuList[0]);
                        }
                    }
                    #endregion
                }
            }
            return;
        }

        /// <summary>
        /// checks if the ant has to get back to the starting city and makes the move (if neccesary)
        /// </summary>
        private void ReturnHomeForNextAgent()
        {
            if ((this.tabuList.Count -1- this.tabuList.LastIndexOf(tabuList[0])) >= citysPerAgent)
            {
                if ((this.cityOverflowVRP > 0) && ((this.tabuList.Count - 1 - this.tabuList.LastIndexOf(tabuList[0])) < (citysPerAgent + 1)))
                {
                    this.cityOverflowVRP -= 1;
                }
                else
                {
                    this.tabuList.Add(tabuList[0]);
                }
            }
        }

        #endregion

        #region calculation and pheromones

        /// <summary>
        /// calculates the route length of the ant depending on the tabu list
        /// </summary>
        /// <returns>the length of the ant's route</returns>
        internal double CalculateRouteLength()
        {
            double routeLength=0;
            if (this.tabuList.Count == 0)
            {
                return double.MaxValue;
            }
            int lastCity = this.tabuList[tabuList.Count-1];
            foreach (int i in this.tabuList)
            {
                routeLength += AntColony.Instance.distances[i, lastCity];
                lastCity = i;
            }
            this.foundRouteLength = routeLength;
            return routeLength;
        }

        /// <summary>
        /// gives the ant permission to add a pheromone trail on the 
        /// trail matrix in the ant colony, multiplied by the given faktor
        /// (used by elitism to separate the better routes from the rest)
        /// </summary>
        /// <param name="faktor">pheromone multiplier for the route</param>
        internal void AddTrail(double faktor)
        {
            double valueToAdd=0;
            switch (AntColony.Instance.PheromoneTyp)
            {
                case PheromoneType.Length:
                    valueToAdd = (AntColony.Instance.PheromoneFaktor * AntColony.Instance.cityList.Count) / (foundRouteLength * AntColony.Instance.AntNumber);
                    break;
                case PheromoneType.Simple:
                    valueToAdd = AntColony.Instance.PheromoneFaktor / foundRouteLength;
                    break;
                case PheromoneType.Exponential:
                    valueToAdd = AntColony.Instance.PheromoneFaktor * AntColony.Instance.cityList.Count * (AntColony.Instance.BestRouteList[AntColony.Instance.BestRouteList.Count - 1].routeLength / foundRouteLength);
                    break;
                default:
                    break;
            }
            valueToAdd *= faktor;
            if (this.tabuList.Count <2 )
            {
                return;
            }
            int lastCity = this.tabuList[tabuList.Count - 1];
            foreach (int i in this.tabuList)
            {
                AntColony.Instance.trail[i, lastCity] += valueToAdd;
                AntColony.Instance.trail[lastCity, i] += valueToAdd;
                lastCity = i;
            }
        }

        #endregion

        #endregion
    }
}
