﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ACO
{
    #region Types Enums

    public enum ProblemType
    {
        TSP = 0,
        VRP = 1,
        error = 2,
    }

    public enum PheromoneType
    {
        Simple = 0,
        Length = 1,
        Exponential = 2,
    }

    #endregion

    public class AntColony
    {
        #region Fields and Variables

        #region Algorythm Type variables
        private AlgType algType;     
        private PheromoneType pheromoneType;
        #endregion
     
        #region Parameters
        private int antNumber;
        private int agentNumber;
        private double alpha;
        private double beta;
        private double rho;
        private double pheromoneFaktor;
        #endregion

        #region Problem
        private ProblemType problemType;
        internal List<City> cityList; 
        internal double[,] visibility;
        internal double[,] distances;
        #endregion

        #region Results
        private List<double> averageRouteLength;
        private List<int> bestRouteCityRow;
        private List<ResultFound> bestRouteList;
        #endregion

        #region Work Variables
        internal Random randomizer; 
        public double[,] trail; 
        internal int cyclusCounter;
        internal List<Ant> ants;
        internal DateTime startedAt;
        internal DateTime endAt;
        private StopConditions stopCondition;
        private static AntColony instance = null;
        #endregion

        #endregion

        #region Methods

        #region Sets, Resets and Initializing

        /// <summary>
        /// private colony constructor, sets the default parameter values and 
        /// initializes lists and arrays
        /// </summary>
        private AntColony()
        {
            //types
            this.algType = new AlgTypeACO(false, false, false, false);
            this.pheromoneType = PheromoneType.Length;
            //parameters
            antNumber = 10;
            agentNumber = 1;
            alpha = 1;
            beta = 1;
            rho = 0.7;
            pheromoneFaktor = 10;
            //problem
            this.problemType = ProblemType.TSP;
            this.cityList = new List<City>();
            this.visibility = new double[1, 1];
            //results
            this.averageRouteLength = new List<double>();
            this.bestRouteCityRow = new List<int>();
            this.bestRouteList = new List<ResultFound>();
            //work vars
            this.randomizer = new Random((int)DateTime.Now.Ticks);
            this.trail = new double[1, 1];
            cyclusCounter = 0;
            ants = new List<Ant>();
            startedAt = DateTime.Now;
            endAt = DateTime.Now;
        }

        /// <summary>
        /// deletes all optimization results found from the colony
        /// </summary>
        public void Reset()
        {
            this.cyclusCounter = 0;
            this.SetTrail(this.pheromoneFaktor);
            this.ants.Clear();
            this.averageRouteLength = new List<double>();
            this.bestRouteCityRow = new List<int>();
            this.bestRouteList = new List<ResultFound>();
        }

        /// <summary>
        /// creates the trail matrix as a new double[,] matrix and fills it with the given value
        /// </summary>
        /// <param name="value">value of all trail paths</param>
        internal void SetTrail(double value)
        {
            this.trail = new double[this.cityList.Count,this.cityList.Count];
            int i = 0;
            int j = 0;
            for (i = 0; i < cityList.Count; i++)
            {
                for (j = 0; j < cityList.Count; j++)
                {
                    this.trail[i, j] = value;
                };
            };
        }

        /// <summary>
        /// creates visibility (and distance) matrixes depending on the loaded problem,
        /// calculates (and saves) the distances between the loaded citys and converts them to visibility, too.
        /// </summary>
        private void SetVisibility()
        {
            this.visibility = new double[this.cityList.Count, this.cityList.Count];
            this.distances = new double[this.cityList.Count, this.cityList.Count];
            for (int i = 0; i < this.cityList.Count; i++)
            {
                for (int j = 0; j < this.cityList.Count; j++)
                {
                    this.distances[i, j] = Math.Sqrt(Math.Pow(AntColony.Instance.cityList[i].X - AntColony.Instance.cityList[j].X, 2) +
                                                   Math.Pow(AntColony.Instance.cityList[i].Y - AntColony.Instance.cityList[j].Y, 2));
                    this.visibility[i, j] = 1.0 / (distances[i, j]);
                }
            }
        }

        #endregion

        #region Input Output Problem

        /// <summary>
        /// adds a new city to the problem. 
        /// Also refreshes the trail, distance and visibility matrixes
        /// called by the main user interface by clicking on the city map
        /// </summary>
        /// <param name="X">x coordinate of the new city</param>
        /// <param name="Y">y coordinate of the new city</param>
        public void AddCity(double X, double Y)
        {
            this.cityList.Add(new City(X, Y));
            this.Reset();
            this.SetVisibility();
            return;
        }

        /// <summary>
        /// removes all citys from the ant colony
        /// </summary>
        public void ClearProblem()
        {
            this.cityList.Clear();
            this.Reset();
            this.SetVisibility();
            return;
        }

        /// <summary>
        /// loads a problem into the ant colony from a file given by the filePath
        /// </summary>
        /// <param name="filePath">path to the file from wich to read the problem</param>
        public void ReadProblemFromFile(string filePath)
        {
            this.ClearProblem();
            OpenProblemClass.Instance.ReadProblemFromFile(filePath);
            this.Reset();
            this.SetVisibility();
            return;
        }

        /// <summary>
        /// creates a new problem with 3 circles
        /// </summary>
        /// <param name="numberOfCitys">preferred number of citys (can deviate a bit)</param>
        public void CreateProblemCircle(int numberOfCitys)
        {
            this.ClearProblem();
            OpenProblemClass.Instance.MakeCircle(numberOfCitys);
            this.Reset();
            this.SetVisibility();
        }

        /// <summary>
        /// creates a mesh-like problem with the given number of citys
        /// (rounds the number to a nearby square of a number. example: 50 citys -> 7*7 mesh)
        /// </summary>
        /// <param name="numberOfCitys">preffered number of citys</param>
        public void CreateProblemMesh(int numberOfCitys)
        {
            this.ClearProblem();
            OpenProblemClass.Instance.MakeCube(numberOfCitys);
            this.Reset();
            this.SetVisibility();
        }

        /// <summary>
        /// creates the ACO--letters-looking problem
        /// </summary>
        public void CreateAcoProblem()
        {
            this.ClearProblem();
            OpenProblemClass.Instance.MakeACOLetters();
            this.Reset();
            this.SetVisibility();
        }

        /// <summary>
        /// creates a problem with "numberOfCitys" citys, randomly created (the coordinates are random)
        /// </summary>
        /// <param name="numberOfCitys">number of random citys</param>
        public void CreateRandomProblem(int numberOfCitys)
        {
            this.ClearProblem();
            OpenProblemClass.Instance.MakeRandom(numberOfCitys);
            this.Reset();
            this.SetVisibility();
        }

        /// <summary>
        /// calls the SaveProblemClass to save the loaded problem into a file given by the filePath
        /// </summary>
        /// <param name="filePath">path to the file to save the problem to</param>
        public void SaveProblem(string filePath)
        {
            if (this.cityList.Count > 0)
            {
                SaveProblemClass.Instance.SaveProblem(filePath);
            }
            return;
        }

        #endregion

        #region Algorithm

        /// <summary>
        /// checks all ats' routes, saves the best route and updates the 
        /// information on the average found route length
        /// </summary>
        private void CheckBestRoutes()
        {
            double best = double.MaxValue;
            if (this.bestRouteList.Count > 0)
            {
                best = this.bestRouteList[bestRouteList.Count - 1].routeLength;
            }
            int bestAnt=-1;
            double sum = 0;
            int antsInCycle = 0;
            foreach (Ant a in this.ants)
            {
                double length = a.CalculateRouteLength();
                if (length < double.MaxValue)
                {
                    antsInCycle++;
                    sum += length;
                    if (length < best)
                    {
                        best = length;
                        bestAnt = ants.IndexOf(a);
                    }
                }
            }
            if (antsInCycle > 0)
            {
                this.averageRouteLength.Add(sum / antsInCycle);
            }
            if (bestAnt != -1)
            {
                bestRouteList.Add(new ResultFound(best, this.cyclusCounter, DateTime.Now - this.startedAt));
                this.bestRouteCityRow.Clear();
                foreach (int i in this.ants[bestAnt].tabuList)
                {
                    this.bestRouteCityRow.Add(i);
                }
            }
            return;
        }

        /// <summary>
        /// calls all ants to add their share of the pheromone to the trail matrix
        /// </summary>
        internal void AddTrail()
        {
            foreach (Ant a in this.ants)
            {
                a.AddTrail(1);
            }
        }

        /// <summary>
        /// calls all ants to make their cycle
        /// </summary>
        private void MakeCyclus()
        {
            foreach (Ant a in this.ants)
            {
                if (this.stopCondition.IsStopConditionTrue())
                {
                    return;
                }
                a.makeCycle();
            }
            return;
        }

        /// <summary>
        /// multiplies all trail values with the given vaporization faktor (rho)
        /// </summary>
        private void Vaporize()
        {
            if (this.algType.AsString() == "Greedy")
            {
                return;
            }
            for (int i = 0; i < this.cityList.Count ; i++)
            {
                for (int j = 0; j < this.cityList.Count; j++)
                {
                    this.trail[i, j] = this.rho * this.trail[i, j];
                };
            };
            return;
        }

        /// <summary>
        /// resets all ants in the colony to make them ready for a new cycle
        /// </summary>
        private void ResetAnts()
        {
            foreach (Ant a in this.ants)
            {
                a.ResetAnt();
            }
        }

        /// <summary>
        /// starts the optimization with the predefined parameters, stop conditions,
        /// algorithm and pheromone types
        /// </summary>
        public void Start()
        {
            if (this.cityList.Count < 2)
            { return; }

            this.startedAt = DateTime.Now;
            if (this.ants.Count != this.antNumber)
            {
                this.ants.Clear();
                for (int i = 0; i < this.antNumber; i++)
                {
                    ants.Add(new Ant());
                }
            }
            while (!this.stopCondition.IsStopConditionTrue())
            {
                this.MakeCyclus();
                this.cyclusCounter++;
                this.CheckBestRoutes();
                this.Vaporize();
                this.algType.DoVersionExtras();
                this.ResetAnts();
            }
            this.endAt = DateTime.Now;
        }

        /// <summary>
        /// defines the algorithm type, pheromone type and stop 
        /// conditions for the colony and calls the regular 
        /// (not parameterized) version of the Start function "Start()"
        /// </summary>
        /// <param name="type">new algorithm type</param>
        /// <param name="phtype">new pheromone type</param>
        /// <param name="stopMe"> new stop conditions</param>
        public void Start(AlgType type, PheromoneType phtype, StopConditions stopMe)
        {
            this.algType = type;
            this.PheromoneTyp = phtype;
            this.StopCondition = stopMe;
            this.Start();
        }

        #endregion

        #endregion

        #region Properties

        /// <summary>
        /// the Singleton design pattern, to insure the existance of only one ant colony in one program
        /// </summary>
        public static AntColony Instance
        {
            get
            {
                if (AntColony.instance == null)
                {
                    AntColony.instance = new AntColony();
                }
                return AntColony.instance;
            }
        }

        /// <summary>
        /// number of ants (min 1, max 9999)
        /// </summary>
        public int AntNumber
        {
            get
            {
                return this.antNumber;
            }
            set
            {
                if ((value > 0) && (value < 10000))
                {
                    this.antNumber = value;
                }
            }
        }

        /// <summary>
        /// number of agents in a VRP (min 1, max 999)
        /// </summary>
        public int AgentNumber
        {
            get
            {
                return this.agentNumber;
            }
            set
            {
                if ((value > 0) && (value < 1000))
                {
                    this.agentNumber = value;
                }
            }
        }

        /// <summary>
        /// alpha - trail importance (min 0, max 1000)
        /// </summary>
        public double Alpha
        {
            get
            {
                return this.alpha;
            }
            set
            {
                if ((value >= 0) && (value <= 1000))
                {
                    this.alpha = value;
                }
            }
        }

        /// <summary>
        /// beta - visibility (distance) importance (min 0, max 1000)
        /// </summary>
        public double Beta
        {
            get
            {
                return this.beta;
            }
            set
            {
                if ((value >= 0) && (value <= 1000))
                {
                    this.beta = value;
                }
            }
        }

        /// <summary>
        /// rho - vaporization faktor (min 0, max 1)
        /// </summary>
        public double Rho
        {
            get
            {
                return this.rho;
            }
            set
            {
                if ((value >= 0) && (value <=1))
                {
                    this.rho = value;
                }
            }
        }

        /// <summary>
        /// pheromone faktor (min 0, max double.MaxValue)
        /// </summary>
        public double PheromoneFaktor
        {
            get
            {
                return this.pheromoneFaktor;
            }
            set
            {
                if (value > 0)
                {
                    this.pheromoneFaktor = value;
                }
            }
        }

        /// <summary>
        /// get the list of average route lengths
        /// </summary>
        public List<double> AverageRouteLength
        {
            get
            {
                return this.averageRouteLength;
            }
        }
   
        /// <summary>
        /// get the list of city indexes of the best found route
        /// </summary>
        public List<int> BestRouteCityRow
        {
            get
            {
                return this.bestRouteCityRow;
            }
        }

        /// <summary>
        /// get the list of best found routes (when they were found and how long they are)
        /// </summary>
        public List<ResultFound> BestRouteList
        {
            get
            {
                return this.bestRouteList;
            }
        }

        /// <summary>
        /// gets or sets the Problem type (VRP or TSP)
        /// </summary>
        public ProblemType ProblemTyp
        {
            get
            {
                return this.problemType;
            }
            set
            {
                if (value != ProblemType.error)
                {
                    this.problemType = value;
                }
            }
        }

        /// <summary>
        /// gets or sets the phromone type
        /// </summary>
        public PheromoneType PheromoneTyp
        {
            get
            {
                return this.pheromoneType;
            }
            set
            {
                this.pheromoneType = value;
            }
        }

        /// <summary>
        /// gets or sets the ACO algorithm type
        /// </summary>
        public AlgType AlgType
        {
            get
            {
                return this.algType;
            }
            set
            {
                this.algType = value;
            }
        }

        /// <summary>
        /// gets or sets the stop conditions for one optimization
        /// </summary>
        public StopConditions StopCondition
        {
            get
            {
                return this.stopCondition;
            }
            set
            {
                this.stopCondition = value;
            }
        }

        /// <summary>
        /// duration of the last run optimization
        /// </summary>
        public TimeSpan LastOptimizationDuration
        {
            get
            {
                return (this.endAt - this.startedAt);
            }
        }

        /// <summary>
        /// list of all citys in the loaded problem
        /// </summary>
        public List<City> CityList
        {
            get
            {
                return this.cityList;
            }
        }

        /// <summary>
        /// gets the number of cycles run in the last optimization
        /// </summary>
        public int CyclusCounter
        {
            get
            {
                return this.cyclusCounter;
            }
        }

        #endregion
    }
}
