﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Shapes;
using Microsoft.Win32;
using System.IO;
using ACO;

namespace WPF_Window
{
    /// <summary>
    /// Interaction logic for Series.xaml
    /// </summary>
    public partial class Series : Window
    {
        #region fields and attributes

        private DateTime startTime;
        private TimeSpan SeriesDuration;
        private AlgType at;
        private StopConditions sc;
        private PheromoneType pt;
        private double testedParameterValue;
        private double LimitValue;
        private double StepValue;

        #endregion

        #region methods

        /// <summary>
        /// constructor, needs the ACOType from the original window to read it properly
        /// </summary>
        /// <param name="aType">ACOType to use in the optimization series</param>
        public Series(AlgType aType)
        {
            InitializeComponent();
            this.at = aType;
        }

        private void ParameterComboBox_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (this.ParameterComboBox.SelectedIndex == 8) 
            {
                //if ACO version is selected set for the number of different versions
                //set the start and end values to indicate the number of ACOtype possibilities
                this.StartValueBox.Text = "0";
                this.StepValueBox.Text = "1";
                this.LimitValueBox.Text = "17";
            }
        }

        /// <summary>
        /// sets the selected parameters value to the given number, in case of ACOType 
        /// changing calls the SetACOType function
        /// </summary>
        /// <param name="Value">the new value for the selected parameter</param>
        private void AdjustTestingParameter(double Value)
        {
            switch (this.ParameterComboBox.SelectedIndex)
            {
                case 0: 
                //agent number
                    AntColony.Instance.AgentNumber = (int)Math.Round(Value);
                    break;
                case 1: 
                //ant number
                    AntColony.Instance.AntNumber = (int)Math.Round(Value);
                    break;
                case 2: 
                //alpha
                    AntColony.Instance.Alpha = Value;
                    break;
                case 3: 
                //beta
                    AntColony.Instance.Beta = Value;
                    break;
                case 4: 
                //rho
                    AntColony.Instance.Rho = Value;
                    break;
                case 5: 
                //pheromone factor
                    AntColony.Instance.PheromoneFaktor = Value;
                    break;
                case 6: 
                //cycle number
                    this.sc.cyclusNumber = (int)Math.Round(Value);
                    break;
                case 7: 
                //Time given
                    this.sc.optimizationTime = new TimeSpan(0, 0, (int)Math.Round(Value));
                    break;
                case 8: 
                //ACO version
                    this.SetACOType((int)Math.Round(Value));
                    break;
                default:
                    //unknown selection
                    MessageBox.Show("unknown selection in combo box!");
                    break;
            }
        }

        /// <summary>
        /// sets the known ACOType acording to the given version of the algorithm
        /// </summary>
        /// <param name="version">version of the algorithm (min 0, max 16; cycles through bigger numbers)</param>
        private void SetACOType(int version)
        {
            switch (version % 17)
            {
                case 0: 
                //Greedy
                    this.at = new AlgTypeGreedy();
                    break;
                case 1: 
                //all OFF
                    this.at = new AlgTypeACO();
                    break;
                case 2:
                    //minmax
                    this.at = new AlgTypeACO(false,false,true,false);
                    break;
                case 3:
                    //system
                    this.at = new AlgTypeACO(false, false, false, true);
                    break;
                case 4:
                    //FootStepping
                    this.at = new AlgTypeACO(false, true, false, false);
                    break;
                case 5:
                    //elitism
                    this.at = new AlgTypeACO(true, false, false, false);
                    break;
                case 6:
                    //minmax+system
                    this.at = new AlgTypeACO(false, false, true, true);
                    break;
                case 7:
                    //minmax+FS
                    this.at = new AlgTypeACO(false, true, true, false);
                    break;
                case 8:
                    //minmax+elitism
                    this.at = new AlgTypeACO(true, false, true, false);
                    break;
                case 9:
                    //system+FS
                    this.at = new AlgTypeACO(false, true, false, true);
                    break;
                case 10:
                    //system+elitism
                    this.at = new AlgTypeACO(true, false, false, true);
                    break;
                case 11:
                    //FS+elitism
                    this.at = new AlgTypeACO(true, true, false, false);
                    break;
                case 12:
                    //all -elitism
                    this.at = new AlgTypeACO(false, true, true, true);
                    break;
                case 13:
                    //all -footstepping
                    this.at = new AlgTypeACO(true, false, true, true);
                    break;
                case 14:
                    //all -system
                    this.at = new AlgTypeACO(true, true, true, false);
                    break;
                case 15:
                    //all -minmax
                    this.at = new AlgTypeACO(true, true, false, true);
                    break;
                case 16:
                    //all ON
                    this.at = new AlgTypeACO(true, true, true, true);
                    break;
                default:
                    break;
            }
        }
        
        /// <summary>
        /// checks if the optimization series has reached any of the stop conditions
        /// </summary>
        /// <returns></returns>
        private Boolean IsStopConditionReached()
        {
            if ((bool)this.StopGlobalTimeCheckBox.IsChecked == true)
            {
                if (DateTime.Now - this.startTime > this.SeriesDuration)
                {
                    return true;
                }
            }
            if (this.testedParameterValue > this.LimitValue)
            {
                return true;
            }
            return false;
        }

        #endregion

        #region button clicks

        private void StartSeriesButton_Click(object sender, RoutedEventArgs e)
        {
            SaveFileDialog sf = new SaveFileDialog();
            sf.DefaultExt = "txt";
            sf.AddExtension = true;
            sf.ShowDialog();
            if (sf.FileName!="")
            {
                this.SeriesOk(sf.FileName);
            }
        }

        private void SeriesOk(string filePath)
        {
            this.startTime = DateTime.Now;
            this.pt = new PheromoneType();
            this.sc = new StopConditions((bool)this.StopOnTimeCheckBox.IsChecked, (bool)this.StopOnCyclusCheckBox.IsChecked, (bool)this.StopMeFileCheckBox.IsChecked);
            #region reading text boxes
            try
            {
                this.SeriesDuration = new TimeSpan(0, 0, int.Parse(this.GlobalTimeLimitBox.Text));
                this.testedParameterValue = double.Parse(this.StartValueBox.Text);
                this.LimitValue = double.Parse(this.LimitValueBox.Text);
                this.StepValue = double.Parse(this.StepValueBox.Text);
            }
            catch
            {
                MessageBox.Show("error parsing double values from text boxes!");
            }
            try
            {
                if ((bool)this.StopOnTimeCheckBox.IsChecked)
                {
                    sc.optimizationTime = new TimeSpan(0, 0, int.Parse(this.StopTimeLimitBox.Text));
                }
                if ((bool)this.StopOnCyclusCheckBox.IsChecked)
                {
                    sc.cyclusNumber = int.Parse(this.StopCyclusLimitBox.Text);
                }
            }
            catch
            {
                MessageBox.Show("Error in reading stop condition values!\ncheck for characters and/or floating points in integer values");
            }
            #endregion
            if ((sc.timeSpan == false) && ((bool)this.StopGlobalTimeCheckBox.IsChecked == true))
            {
                sc.timeSpan = true;
                sc.optimizationTime = this.SeriesDuration;
            }
            SaveDataClass.Instance.ResetSaveString();
            while (!this.IsStopConditionReached())
            {
                AntColony.Instance.Reset();
                this.AdjustTestingParameter(this.testedParameterValue);
                if ((bool)this.StopGlobalTimeCheckBox.IsChecked == true)
                {
                    TimeSpan GlobalMax = this.SeriesDuration - (DateTime.Now - this.startTime);
                    if (GlobalMax < sc.optimizationTime)
                    {
                        sc.optimizationTime = GlobalMax;
                    }
                }
                AntColony.Instance.Start(at, pt, sc);
                SaveDataClass.Instance.AddForSeries();
                this.testedParameterValue += this.StepValue;
            }
            if (SaveDataClass.Instance.Save(filePath) == false)
            {
                MessageBox.Show("error while saving");
            }
        }

        private void CancelButton_Click(object sender, RoutedEventArgs e)
        {
            this.Close();
        }

        private void EstimateButton_Click(object sender, RoutedEventArgs e)
        {
            double all = double.Parse(this.StopTimeLimitBox.Text);
            if ((bool)this.StopOnCyclusCheckBox.IsChecked)
            {
                int n = AntColony.Instance.CityList.Count;
                int m = AntColony.Instance.AntNumber;
                double singleTime = 0;
                if (this.at.AsString() == "Greedy")
                {
                    singleTime = 2 / (2000 * 2500);
                }
                else
                {
                    singleTime = 3.1 / (2000 * 2500);
                }
                all = n * n * m * singleTime * (int.Parse(this.StopCyclusLimitBox.Text));
            }
            if ((all > double.Parse(this.StopTimeLimitBox.Text)) && ((bool)this.StopOnTimeCheckBox.IsChecked))
            {
                all = double.Parse(this.StopTimeLimitBox.Text);
            }
            int repeating = (int)Math.Round((double.Parse(this.LimitValueBox.Text) - double.Parse(this.StartValueBox.Text)) / double.Parse(this.StepValueBox.Text));
            if (repeating < 1)
            {
                all = 0;
            }
            else
            {
                all *= repeating;
            }
            if (((bool)this.StopGlobalTimeCheckBox.IsChecked == true) && (double.Parse(this.GlobalTimeLimitBox.Text) < all))
            {
                all = double.Parse(this.GlobalTimeLimitBox.Text);
            }
            this.EstimatedTimeBox.Text = (new TimeSpan(0, 0, (int)Math.Round(all))).ToString();
        }

        #endregion

    }
}
